@@ -175,6 +175,13 @@ void FilterCriteriaModelPrivate::rebuildModel()
                                                        mRootItem);
         mRootItem->appendChild(parent);
         QVector<QString> units = JournaldHelper::queryUnique(mJournal, JournaldHelper::Field::_SYSTEMD_UNIT);
+        units.erase(std::remove_if(std::begin(units),
+                                   std::end(units),
+                                   [](const QString &unit) {
+                                       return unit.startsWith(QLatin1String("systemd-coredump@"))
+                                           || unit.startsWith(QLatin1String("drkonqi-coredump-processor@"));
+                                   }),
+                    std::cend(units));
         std::sort(std::begin(units), std::end(units), [](const QString &a, const QString &b) {
             return QString::compare(a, b, Qt::CaseInsensitive) <= 0;
         });
