@@ -33,6 +33,92 @@ void JournaldViewModelPrivate::resetJournal()
 
     qCDebug(KJOURNALDLIB_FILTERTRACE) << "flush_matches()";
 
+    auto addConjunction = [](sd_journal *journal) -> void {
+        int result{0};
+        result = sd_journal_add_conjunction(journal);
+        if (result < 0) {
+            qCCritical(KJOURNALDLIB_FILTERTRACE).nospace() << "add_conjunction returned error";
+        }
+        Q_ASSERT(result >= 0);
+    };
+
+    auto addDisjunction = [](sd_journal *journal) -> void {
+        int result{0};
+        result = sd_journal_add_disjunction(journal);
+        if (result < 0) {
+            qCCritical(KJOURNALDLIB_FILTERTRACE).nospace() << "add_disjunction returned error";
+        }
+        Q_ASSERT(result >= 0);
+    };
+
+    auto addMatchesBootFilter = [](sd_journal *journal, const QStringList &boots) -> void {
+        int result{0};
+        for (const QString &boot : boots) {
+            QString filterExpression = QLatin1String("_BOOT_ID=") + boot;
+            result = sd_journal_add_match(journal, filterExpression.toUtf8().constData(), 0);
+            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
+            if (result < 0) {
+                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
+            }
+        }
+    };
+
+    auto addMatchesPriorityFilter = [](sd_journal *journal, std::optional<quint8> priorityLimit) -> void {
+        int result{0};
+        if (priorityLimit.has_value()) {
+            for (int i = 0; i <= *priorityLimit; ++i) {
+                QString filterExpression = QLatin1String("PRIORITY=") + QString::number(i);
+                result = sd_journal_add_match(journal, filterExpression.toUtf8().constData(), 0);
+                qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
+                if (result < 0) {
+                    qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
+                }
+            }
+            qCDebug(KJOURNALDLIB_GENERAL) << "Use priority filter level:" << *priorityLimit;
+        } else {
+            qCDebug(KJOURNALDLIB_GENERAL) << "Skip setting priority filter";
+        }
+    };
+
+    auto addMatchesUnitFilter = [](sd_journal *journal, const QStringList &units) -> void {
+        int result{0};
+        for (const QString &unit : units) {
+            QString filterExpression = QLatin1String("_SYSTEMD_UNIT=") + unit;
+            result = sd_journal_add_match(journal, filterExpression.toUtf8().constData(), 0);
+            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
+            if (result < 0) {
+                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
+            }
+        }
+    };
+
+    auto addMatchesExeFilter = [](sd_journal *journal, const QStringList &exes) -> void {
+        int result{0};
+        for (const QString &executable : exes) {
+            QString filterExpression = QLatin1String("_EXE=") + executable;
+            result = sd_journal_add_match(journal, filterExpression.toUtf8().constData(), 0);
+            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
+            if (result < 0) {
+                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
+            }
+        }
+    };
+
+    auto addMatchesTransportFilter = [](sd_journal *journal, const QStringList &transports) -> void {
+        int result{0};
+        for (const QString &transport : transports) {
+            QString filterExpression = QLatin1String("_TRANSPORT=") + transport;
+            result = sd_journal_add_match(journal, filterExpression.toUtf8().constData(), 0);
+            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
+            if (result < 0) {
+                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
+            }
+        }
+    };
+
+    const QStringList kernelTransports{QLatin1String("audit"), QLatin1String("driver"), QLatin1String("kernel")};
+    const QStringList nonKernelTransports{QLatin1String("syslog"), QLatin1String("journal"), QLatin1String("stdout")};
+
     // filter construction:
     // The Journald API does not provide arbitrary logical phrases, but a 4 level syntax,
     // see: https://www.freedesktop.org/software/systemd/man/sd_journal_add_match.html
@@ -44,91 +130,52 @@ void JournaldViewModelPrivate::resetJournal()
     // The following boolean expression is created as follow for kernel transport option:
     //     (boot=123 OR boot=...)
     //     AND (priority=1 OR priority=...)
-    //     AND (transport=kernel) OR (unit_1 OR unit_2 OR ...) OR (exe=x OR exe=y OR ...)
-    // And for non-kernel transport option:
+    //     AND (transport=kernel)
+    // OR
     //     (boot=123 OR boot=...)
     //     AND (priority=1 OR priority=...)
-    //     AND (transport=not-kernel)
-    //     AND (unit_1 OR unit_2 OR ...) OR (exe=x OR exe=y OR ...)
+    //     AND (unit_1 OR unit_2 OR ...)
+    // OR
+    //     (boot=123 OR boot=...)
+    //     AND (priority=1 OR priority=...)
+    //     AND (exe=x OR exe=y OR ...)
 
-    // filter boots
-    for (const QString &boot : qAsConst(mBootFilter)) {
-        QString filterExpression = QLatin1String("_BOOT_ID=") + boot;
-        result = sd_journal_add_match(mJournal->sdJournal(), filterExpression.toLocal8Bit().constData(), 0);
-        qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
-        if (result < 0) {
-            qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
+    bool clauseAdded{false};
+    // kernel filter is special in the sense that thouse message only shall be added
+    // and in the absense of different category filters, a filtering of the correct
+    // transport layer must be applied
+    if (mFilter.areKernelMessagesEnabled()) {
+        clauseAdded = true;
+        addMatchesBootFilter(mJournal->sdJournal(), mFilter.bootFilter());
+        addMatchesPriorityFilter(mJournal->sdJournal(), mFilter.priorityFilter());
+        QStringList transportFilter = kernelTransports;
+        if (mFilter.systemdUnitFilter().empty() && mFilter.exeFilter().empty()) {
+            transportFilter.append(nonKernelTransports);
         }
+        addMatchesTransportFilter(mJournal->sdJournal(), transportFilter);
+    } else if (mFilter.systemdUnitFilter().empty() && mFilter.exeFilter().empty()) {
+        clauseAdded = true;
+        addMatchesBootFilter(mJournal->sdJournal(), mFilter.bootFilter());
+        addMatchesPriorityFilter(mJournal->sdJournal(), mFilter.priorityFilter());
+        addMatchesTransportFilter(mJournal->sdJournal(), nonKernelTransports);
     }
-    if (mPriorityFilter.has_value()) {
-        for (int i = 0; i <= mPriorityFilter; ++i) {
-            QString filterExpression = QLatin1String("PRIORITY=") + QString::number(i);
-            result = sd_journal_add_match(mJournal->sdJournal(), filterExpression.toLocal8Bit().constData(), 0);
-            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
-            if (result < 0) {
-                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
-            }
-        }
-        qCDebug(KJOURNALDLIB_GENERAL) << "Use priority filter level:" << mPriorityFilter.value();
-    } else {
-        qCDebug(KJOURNALDLIB_GENERAL) << "Skip setting priority filter";
+    if (clauseAdded && !mFilter.systemdUnitFilter().empty()) {
+        addDisjunction(mJournal->sdJournal());
+        clauseAdded = false;
     }
-
-    // boot and priority filter shall always be enforced
-    result = sd_journal_add_conjunction(mJournal->sdJournal());
-    qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_conjunction()";
-    Q_ASSERT(result >= 0);
-
-    // see journal-fields documentation regarding list of valid transports
-    // note: in case of kernel messages being activated, this filter automatically activates all kernel transport
-    //       because kernel output will not match any further service/exe filter
-    QStringList kernelTransports{QLatin1String("audit"), QLatin1String("driver"), QLatin1String("kernel")};
-    QStringList nonKernelTransports{QLatin1String("syslog"), QLatin1String("journal"), QLatin1String("stdout")};
-    if (mShowKernelMessages) {
-        for (const QString &transport : kernelTransports) {
-            QString filterExpression = QLatin1String("_TRANSPORT=") + transport;
-            result = sd_journal_add_match(mJournal->sdJournal(), filterExpression.toLocal8Bit().constData(), 0);
-            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
-            if (result < 0) {
-                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
-            }
-        }
-        result = sd_journal_add_disjunction(mJournal->sdJournal());
-        Q_ASSERT(result >= 0);
-    } else {
-        for (const QString &transport : nonKernelTransports) {
-            QString filterExpression = QLatin1String("_TRANSPORT=") + transport;
-            result = sd_journal_add_match(mJournal->sdJournal(), filterExpression.toLocal8Bit().constData(), 0);
-            qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
-            if (result < 0) {
-                qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
-            }
-        }
-        result = sd_journal_add_conjunction(mJournal->sdJournal());
-        Q_ASSERT(result >= 0);
+    if (!mFilter.systemdUnitFilter().empty()) {
+        clauseAdded = true;
+        addMatchesBootFilter(mJournal->sdJournal(), mFilter.bootFilter());
+        addMatchesPriorityFilter(mJournal->sdJournal(), mFilter.priorityFilter());
+        addMatchesUnitFilter(mJournal->sdJournal(), mFilter.systemdUnitFilter());
     }
-
-    // filter units
-    for (const QString &unit : qAsConst(mSystemdUnitFilter)) {
-        QString filterExpression = QLatin1String("_SYSTEMD_UNIT=") + unit;
-        result = sd_journal_add_match(mJournal->sdJournal(), filterExpression.toLocal8Bit().constData(), 0);
-        qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
-        if (result < 0) {
-            qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
-        }
+    if (clauseAdded && !mFilter.exeFilter().empty()) {
+        addDisjunction(mJournal->sdJournal());
     }
-
-    result = sd_journal_add_disjunction(mJournal->sdJournal());
-    Q_ASSERT(result >= 0);
-
-    // filter executable
-    for (const QString &executable : qAsConst(mExeFilter)) {
-        QString filterExpression = QLatin1String("_EXE=") + executable;
-        result = sd_journal_add_match(mJournal->sdJournal(), filterExpression.toLocal8Bit().constData(), 0);
-        qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "add_match(" << filterExpression << ")";
-        if (result < 0) {
-            qCCritical(KJOURNALDLIB_GENERAL) << "Failed to set journal filter:" << strerror(-result) << filterExpression;
-        }
+    if (!mFilter.exeFilter().empty()) {
+        addMatchesBootFilter(mJournal->sdJournal(), mFilter.bootFilter());
+        addMatchesPriorityFilter(mJournal->sdJournal(), mFilter.priorityFilter());
+        addMatchesExeFilter(mJournal->sdJournal(), mFilter.exeFilter());
     }
 
     qCDebug(KJOURNALDLIB_FILTERTRACE).nospace() << "Filter DONE";
@@ -140,6 +187,7 @@ void JournaldViewModelPrivate::resetJournal()
 
 QVector<LogEntry> JournaldViewModelPrivate::readEntries(Direction direction)
 {
+    QString dir = (direction == Direction::TOWARDS_HEAD) ? QStringLiteral("head") : QStringLiteral("tail");
     static QMutex mutex;
     QMutexLocker locker(&mutex);
 
@@ -149,72 +197,38 @@ QVector<LogEntry> JournaldViewModelPrivate::readEntries(Direction direction)
         qCWarning(KJOURNALDLIB_GENERAL) << "Skipping data fetch, no valid journal opened";
         return chunk;
     }
-    if (direction == Direction::TOWARDS_TAIL) {
-        if (mLog.size() > 0) {
-            QString cursor = mLog.last().mCursor;
-            // note: seek cursor does not make it current, but a subsequenct sd_journal_next is required
-            result = sd_journal_seek_cursor(mJournal->sdJournal(), cursor.toLocal8Bit().constData());
-            if (result < 0) {
-                qCWarning(KJOURNALDLIB_GENERAL) << "seeking cursor but could not be found" << strerror(-result);
-            }
-            result = sd_journal_next(mJournal->sdJournal());
-            if (result == 0) {
-                mTailCursorReached = true;
-                return {};
-            }
-            result = sd_journal_test_cursor(mJournal->sdJournal(), cursor.toLocal8Bit().constData());
-            if (result <= 0) {
-                qCCritical(KJOURNALDLIB_GENERAL) << "current position does not match expected cursor:" << cursor;
-                if (result < 0) {
-                    qCCritical(KJOURNALDLIB_GENERAL) << "cursor test failed:" << strerror(-result);
+    if (mLog.size() > 0) {
+        const QString cursor = (direction == Direction::TOWARDS_TAIL) ? mLog.last().mCursor : mLog.first().mCursor;
+        const SeekCursorResult seekResult = seekCursor(cursor);
+        switch (seekResult) {
+        case SeekCursorResult::CURSOR_MADE_CURRENT:
+            if (direction == Direction::TOWARDS_TAIL) {
+                result = sd_journal_next(mJournal->sdJournal());
+                if (result == 0) {
+                    mTailCursorReached = true;
+                    return {};
                 }
-                return {};
-            }
-            // read first entry after cursor
-            result = sd_journal_next(mJournal->sdJournal());
-            if (result == 0) {
-                mTailCursorReached = true;
-                return {};
-            }
-        } else {
-            if (!seekHeadAndMakeCurrent()) {
-                return {};
-            }
-        }
-    } else if (direction == Direction::TOWARDS_HEAD) {
-        if (mLog.size() > 0) {
-            QString cursor = mLog.first().mCursor;
-            result = sd_journal_seek_cursor(mJournal->sdJournal(), cursor.toLocal8Bit().constData());
-            if (result < 0) {
-                qCWarning(KJOURNALDLIB_GENERAL) << "seeking cursor but could not be found" << strerror(-result);
-            }
-            result = sd_journal_previous(mJournal->sdJournal());
-            if (result == 0) {
-                mHeadCursorReached = true;
-                return {};
-            }
-            result = sd_journal_test_cursor(mJournal->sdJournal(), cursor.toLocal8Bit().constData());
-            if (result <= 0) {
-                qCCritical(KJOURNALDLIB_GENERAL) << "current position does not match expected cursor:" << cursor;
-                if (result < 0) {
-                    qCCritical(KJOURNALDLIB_GENERAL) << "cursor test failed:" << strerror(-result);
+            } else if (direction == Direction::TOWARDS_HEAD) {
+                result = sd_journal_previous(mJournal->sdJournal());
+                if (result == 0) {
+                    mHeadCursorReached = true;
+                    return {};
                 }
-                return {};
-            }
-            // read first entry before cursor
-            result = sd_journal_previous(mJournal->sdJournal());
-            if (result == 0) {
-                mHeadCursorReached = true;
-                return {};
-            }
-        } else {
-            if (!seekTailAndMakeCurrent()) {
-                return {};
             }
+            break;
+        case SeekCursorResult::ERROR:
+            qCCritical(KJOURNALDLIB_GENERAL) << "cursor test failed:" << strerror(-result);
+            return {};
+            break;
+        }
+    } else if (mLog.size() == 0 && direction == Direction::TOWARDS_TAIL) {
+        if (!seekHeadAndMakeCurrent()) {
+            return {};
+        }
+    } else if (mLog.size() == 0 && direction == Direction::TOWARDS_HEAD) {
+        if (!seekTailAndMakeCurrent()) {
+            return {};
         }
-    } else {
-        qCCritical(KJOURNALDLIB_GENERAL) << "Jumping into the journal's middle, not supported";
-        return {};
     }
 
     // at this point, the journal is guaranteed to point to the first valid entry
@@ -226,7 +240,7 @@ QVector<LogEntry> JournaldViewModelPrivate::readEntries(Direction direction)
         LogEntry entry;
         result = sd_journal_get_realtime_usec(mJournal->sdJournal(), &time);
         if (result == 0) {
-            entry.mDate.setMSecsSinceEpoch(time / 1000);
+            entry.mDate = QDateTime::fromMSecsSinceEpoch(time / 1000, Qt::UTC);
         }
         sd_id128_t bootId; // currently unused
         result = sd_journal_get_monotonic_usec(mJournal->sdJournal(), &time, &bootId);
@@ -289,6 +303,46 @@ QVector<LogEntry> JournaldViewModelPrivate::readEntries(Direction direction)
     return chunk;
 }
 
+JournaldViewModelPrivate::SeekCursorResult JournaldViewModelPrivate::seekCursor(const QString &cursor)
+{
+    int result{0};
+    // note: seek cursor does not make it current, but a subsequent sd_journal_next is required
+    result = sd_journal_seek_cursor(mJournal->sdJournal(), cursor.toUtf8().constData());
+    if (result < 0) {
+        qCWarning(KJOURNALDLIB_GENERAL) << "seeking cursor but could not be found" << strerror(-result);
+        return SeekCursorResult::ERROR;
+    }
+
+    // make entry current
+    // every result other than 1 is an error, because the cursor is expected to exist
+    result = sd_journal_next(mJournal->sdJournal());
+    if (result != 1) {
+        qCCritical(KJOURNALDLIB_GENERAL) << "seeked entry for cursor could not be made current";
+        return JournaldViewModelPrivate::SeekCursorResult::ERROR;
+    }
+
+    // fallback search logic for problematic versions of systemd: https://github.com/systemd/systemd/issues/31516
+    // note the linear search is very expensive and should be the exception
+    result = sd_journal_test_cursor(mJournal->sdJournal(), cursor.toUtf8().constData());
+    if (result > 0) {
+        return SeekCursorResult::CURSOR_MADE_CURRENT;
+    } else {
+        qCWarning(KJOURNALDLIB_GENERAL) << "current position does not match expected cursor, entering expensive linear search";
+        //(requested, actual):" << cursor << actualCursor;
+        sd_journal_seek_head(mJournal->sdJournal());
+        while (sd_journal_next(mJournal->sdJournal()) > 0) {
+            char *actualCursor{nullptr};
+            const int actualCursorResult = sd_journal_get_cursor(mJournal->sdJournal(), &actualCursor);
+            free(actualCursor);
+            if (actualCursorResult >= 0) {
+                return SeekCursorResult::CURSOR_MADE_CURRENT;
+            }
+        }
+        qCCritical(KJOURNALDLIB_GENERAL) << "even with linear search, the cursor could not be found, giving up";
+    }
+    return SeekCursorResult::ERROR;
+}
+
 bool JournaldViewModelPrivate::seekHeadAndMakeCurrent()
 {
     qCDebug(KJOURNALDLIB_GENERAL) << "seek head and make current";
@@ -365,7 +419,7 @@ bool JournaldViewModel::setJournal(std::unique_ptr<IJournal> journal)
     guardedEndResetModel();
     fetchMoreLogEntries();
     connect(d->mJournal.get(), &IJournal::journalUpdated, this, [=](const QString &bootId) {
-        if (!d->mBootFilter.contains(bootId)) {
+        if (!d->mFilter.bootFilter().contains(bootId)) {
             return;
         }
         if (d->mTailCursorReached) {
@@ -596,102 +650,26 @@ void JournaldViewModel::seekTail()
     guardedEndResetModel();
 }
 
-void JournaldViewModel::setSystemdUnitFilter(const QStringList &systemdUnitFilter)
+void JournaldViewModel::setFilter(const Filter &filter)
 {
+    qCDebug(KJOURNALDLIB_FILTERTRACE) << "setfilter" << filter;
+    // TODO add == operator and skip setting of same filter
     guardedBeginResetModel();
-    d->mSystemdUnitFilter = systemdUnitFilter;
+    d->mFilter = filter;
     d->resetJournal();
     guardedEndResetModel();
     fetchMoreLogEntries();
 }
 
-QStringList JournaldViewModel::systemdUnitFilter() const
+void JournaldViewModel::resetFilter()
 {
-    return d->mSystemdUnitFilter;
+    Filter defaultFilter;
+    setFilter(defaultFilter);
 }
 
-void JournaldViewModel::setBootFilter(const QStringList &bootFilter)
+Filter JournaldViewModel::filter() const
 {
-    if (d->mBootFilter == bootFilter) {
-        return;
-    }
-    guardedBeginResetModel();
-    d->mBootFilter = bootFilter;
-    d->resetJournal();
-    guardedEndResetModel();
-    fetchMoreLogEntries();
-    Q_EMIT bootFilterChanged();
-}
-
-QStringList JournaldViewModel::bootFilter() const
-{
-    return d->mBootFilter;
-}
-
-void JournaldViewModel::setExeFilter(const QStringList &exeFilter)
-{
-    if (d->mExeFilter == exeFilter) {
-        return;
-    }
-    guardedBeginResetModel();
-    d->mExeFilter = exeFilter;
-    d->resetJournal();
-    guardedEndResetModel();
-    fetchMoreLogEntries();
-    Q_EMIT exeFilterChanged();
-}
-
-QStringList JournaldViewModel::exeFilter() const
-{
-    return d->mExeFilter;
-}
-
-void JournaldViewModel::setPriorityFilter(int priority)
-{
-    qCDebug(KJOURNALDLIB_GENERAL) << "Set priority filter to:" << priority;
-    guardedBeginResetModel();
-    if (priority >= 0) {
-        d->mPriorityFilter = priority;
-    } else {
-        d->mPriorityFilter = std::nullopt;
-    }
-    d->resetJournal();
-    guardedEndResetModel();
-    fetchMoreLogEntries();
-    Q_EMIT priorityFilterChanged();
-}
-
-void JournaldViewModel::resetPriorityFilter()
-{
-    guardedBeginResetModel();
-    d->mPriorityFilter.reset();
-    d->resetJournal();
-    guardedEndResetModel();
-    fetchMoreLogEntries();
-    Q_EMIT priorityFilterChanged();
-}
-
-int JournaldViewModel::priorityFilter() const
-{
-    return d->mPriorityFilter.value_or(-1);
-}
-
-void JournaldViewModel::setKernelFilter(bool showKernelMessages)
-{
-    if (d->mShowKernelMessages == showKernelMessages) {
-        return;
-    }
-    guardedBeginResetModel();
-    d->mShowKernelMessages = showKernelMessages;
-    d->resetJournal();
-    guardedEndResetModel();
-    fetchMoreLogEntries();
-    Q_EMIT kernelFilterChanged();
-}
-
-bool JournaldViewModel::isKernelFilterEnabled() const
-{
-    return d->mShowKernelMessages;
+    return d->mFilter;
 }
 
 int JournaldViewModel::search(const QString &searchString, int startRow, Direction direction)
